#include <iostream>
#include <stdlib.h>
#include <stdio.h>
#include <string>
#include <list>
#include <utility>
#include <sys/stat.h>
#include <io.h>
#include <direct.h>

using namespace std;



static  void parseAndExecute(string const &s);

static bool verbose=false;

int main(int argc, char **argv)
{

    static string dashC("-c");





#if 0

    parseAndExecute("if cmd /c true  ; \\ then cmd /c echo --- tests.exe PASSED ; fi"); 

    parseAndExecute("if true ; \\ then  echo yes ; fi");

    parseAndExecute("if [ ! 1 -gt 1  -o 2  -gt 3 ] ; then  echo yes ; fi");

    parseAndExecute(" echo one ; echo two ; echo three ; ");

    parseAndExecute("if [ 7 -gt 8 ] ; then echo yes ; else echo no ; if [ 1 -lt 2 ] ; then echo subyes ; fi ; fi");

    parseAndExecute("if [ -f Makefil1e ] ; then echo makefil1e exists ; fi");

    parseAndExecute("if [ ! -r crap/file1 ] ; then cmd /c echo !!! test4 FAILED ; fi");

#endif

    if(argc < 3 || argc > 1 && dashC != argv[1])
    {
        cerr << "Error:  expected the first argument to be -c" << endl;
        exit(1);
    }

    string cmd;

    parseAndExecute(argv[2]);


}

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct Command
   /// the base class of command handler.  This class is a do nothing command, others
   /// must be derived to execute real work
{
     virtual bool execute() const { return true;  }   // return true if sucessfull execution

     ~Command() {}
};                            


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct RegularCommand
:   public Command
{

    string commandLine_;

    RegularCommand(string const &s)
    : commandLine_(s)
    {
    }


    virtual bool execute() const 
    {
        if(verbose) 
            cout << '(' << commandLine_ << ')' << endl;

        int rv=0;


        char firstWord[256];

        firstWord[0] = 0;

        sscanf(commandLine_.c_str(), "%255.255s", firstWord);

        static string trueToken("true");
        static string falseToken("false");


        if(trueToken == firstWord)
        {
           // let rv remain 0
        }
        else
        if(falseToken == firstWord)
        {
           rv=1; // failed
        }

        else
        {
            rv = system(commandLine_.c_str());
        }

        if(rv == 0)
           return true; // zero means good in system() land

        return false;
    }

    ~RegularCommand()
    {
    }

};

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct CommandList
: public Command
{
    CommandList()
    {
    }

    CommandList(list<Command*> l)
        /// Adopt an existing list of commands and arrange for it to be deleted
        /// at the appropriate time
    : commands_(l)
    {
    }

   virtual ~CommandList()
   {
      list<Command*>::iterator f = commands_.begin(), l = commands_.end();

      while(f != l)
      {
         delete *f++;
      }
   }

   virtual bool execute() const
   {
      
      list<Command*>::const_iterator f = commands_.begin(), l = commands_.end();

      bool rv;

      while(f != l)
      {
          rv = (*f++)->execute();
         
      }

      return rv;  // return the status of the last command
   }


private:
    list<Command*> commands_;

};


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct IfCommand
: public Command
{
    IfCommand(Command *test, Command *ifClause, Command *elseClause, bool invertedLogic)
    :  test_(test)
    ,  ifClause_(ifClause)
    ,  elseClause_(elseClause)
    ,  inverted_(invertedLogic)
    {
    }

    virtual bool execute() const
    {
        bool ok=true;
        
        
        if(test_.get())
           ok = test_->execute();

        if(inverted_)
           ok = !ok;

        if(ok)
        {
           return ifClause_.get() ? ifClause_->execute() : true ;
        }
        else
        {
           return elseClause_.get() ? elseClause_->execute() : true ;
        }
    }

    
    virtual ~IfCommand()
    {
    }


public:
    auto_ptr<Command> test_;
    auto_ptr<Command> ifClause_;
    auto_ptr<Command> elseClause_;
    bool     inverted_;
};

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct TestExpression
    ///
    ///  A representation of test command's syntax.  
    ///
    ///  The execute method returns either true or false to indicate that the
    ///  expression's execution was successful.
    ///
{
    TestExpression() 
    : inverted_(false)
    {
    }

    virtual ~TestExpression() {}  // we typically use self destructing objects or
                                  // auto_ptr<TestExpression> in derived classes,
                                  // but destructors are declared virtual for safety.

    virtual bool execute() const { return true; }

    virtual string stringValue() const { return execute() ? "1" : "0"; }  // most classes don't
                                                                          // need to override this
    void setInverted(bool flag) { inverted_ = flag; }

protected:
    bool inverted_;  // usually false, only set to true when the not operator is used and it is
                     // only allowed at the beginning of a term

};

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct AndExpression
: public TestExpression
{
    AndExpression(TestExpression *lhs, TestExpression *rhs)
    :  lhs_(lhs)
    ,  rhs_(rhs)
    {
    }

    virtual bool execute() const
    {
       if(inverted_ && verbose)
          cout << "!(";
       
       
       bool lhs  = lhs_->execute();

       if(verbose)
          cout << " -a ";

       bool rv = lhs &&  rhs_->execute();

       if(inverted_ && verbose)
          cout << ")";


       if(inverted_)
          rv = !rv;


       return rv;
    }

    virtual ~AndExpression() {} // auto ptrs delete the members


private:
    auto_ptr<TestExpression> lhs_;
    auto_ptr<TestExpression> rhs_;
};

////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct OrExpression
: public TestExpression
{
    OrExpression(TestExpression *lhs, TestExpression *rhs)
    :  lhs_(lhs)
    ,  rhs_(rhs)
    {
    }

    virtual bool execute() const
    {
       if(inverted_ && verbose)
           cout << "!(" ;


       bool lhs = lhs_->execute();

       if(verbose)
           cout << " -o ";


       bool rv = lhs || rhs_->execute();


       if(inverted_ && verbose)
          cout << ")";

       if(inverted_)
          rv = ! rv;

       return rv;
    }



private:
    auto_ptr<TestExpression> lhs_;
    auto_ptr<TestExpression> rhs_;
};


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct StringExpression
: public TestExpression
{
    StringExpression(string const &s)
    :  string_(s)
    {
    }

    virtual bool execute() const
    {


       if(verbose)
          cout << (inverted_ ? "!" : "") << "'" << string_ << "'";

       bool rv =!string_.empty();

       if(inverted_)
          rv = !rv;

       return rv;
    }

    virtual string stringValue() const { return string_; }


private:
    string string_;
};

#ifndef STRINGIFY
#define STRINGIFY(x) #x
#endif



#define DEFINE_TEST_STRING_BINOP(name, operator) \
struct name\
: public TestExpression \
{\
   name(TestExpression *lhs, TestExpression *rhs) \
   : lhs_(lhs) \
   , rhs_(rhs)\
   {\
   }\
\
   virtual ~name() {}\
\
   virtual bool execute() const\
   {\
       if(inverted_ && verbose) cout << "!("; \
       if(verbose) cout << lhs_->stringValue() << STRINGIFY(operator) << rhs_->stringValue() << endl; \
       bool rv =  lhs_->stringValue() operator rhs_->stringValue();\
       if(inverted_ && verbose) cout << ")";\
       if(inverted_)  rv = !rv;\
       return rv;\
   }\
\
private:\
   auto_ptr<TestExpression> lhs_;\
   auto_ptr<TestExpression> rhs_;\
}


DEFINE_TEST_STRING_BINOP(StringEqualExpression, ==);
DEFINE_TEST_STRING_BINOP(StringNotEqualExpression, !=);
DEFINE_TEST_STRING_BINOP(StringGreaterExpression, >);
DEFINE_TEST_STRING_BINOP(StringLessExpression, <);
DEFINE_TEST_STRING_BINOP(StringGreaterEqualExpression, >=);
DEFINE_TEST_STRING_BINOP(StringLessEqualExpression, <=);

#if 0

#define DEFINE_TEST_MATH_BINOP(name, operator) \
struct name\
: public TestExpression \
{\
   name(TestExpression *lhs, TestExpression *rhs) \
   : lhs_(lhs) \
   , rhs_(rhs)\
   {\
   }\
\
   virtual ~name() {}\
\
   virtual bool execute() const\
   {\
       return atoi(lhs_->stringValue().c_str()) operator atoi(rhs_->stringValue().c_str());\
   }\
\
private:\
   auto_ptr<TestExpression> lhs_;\
   auto_ptr<TestExpression> rhs_;\
}


DEFINE_TEST_MATH_BINOP(MathEqualExpression, ==);
DEFINE_TEST_MATH_BINOP(MathNotEqualExpression, !=);
DEFINE_TEST_MATH_BINOP(MathGreaterExpression, >);
DEFINE_TEST_MATH_BINOP(MathLessExpression, <);
DEFINE_TEST_MATH_BINOP(MathGreaterEqualExpression, >=);
DEFINE_TEST_MATH_BINOP(MathLessEqualExpression, <=);
#endif

enum FILEOP
{
    FILEOP_UNSPEC,         // unspecified operation, do not use
    FILEOP_EXISTS,
    FILEOP_READABLE,
    FILEOP_WRITEABLE,
    FILEOP_ISDIR,
    FILEOP_EMPTY,         // non-existent files are empty
    FILEOP_EXE,           // is executable 
};


////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////
struct FileOpExpression
: public TestExpression
{
    FileOpExpression(FILEOP op, string const &s)
    :  string_(s)
    ,  op_(op)
    {
    }

    virtual bool execute() const
    {
       // not yet implemented!

       struct stat fileStatus;

       bool error;

       if(0 != stat(string_.c_str(), &fileStatus))
          error = true;



       switch(op_)
       {
           case FILEOP_EXISTS:
              {

                  if(verbose)
                     cout << (inverted_ ? "!" : "") << "-f " << string_ << ' ';
                  
                  bool rv = !error;
                  
                  if(inverted_)
                     rv = ! rv;
                  
                  
                  return rv;
              }

           case FILEOP_READABLE:
              {
                  if(verbose)
                     cout << (inverted_ ? "!" : "") << "-r " << string_ << ' ';
    
                  bool rv = !error;
    
                  if(inverted_) rv = !rv;
    
                  return rv;
              }

           case FILEOP_WRITEABLE:
              {
                  if(verbose)
                     cout << (inverted_ ? "!" : "") << "-w " << string_ << ' ';
    
                  bool rv= !error;
    
                  if(inverted_)
                     rv=!rv;
    
                  return rv;
              }
           case FILEOP_ISDIR:
              {
                  if(verbose)
                     cout << (inverted_ ? "!" : "") << "-d " << string_ << ' ';
    
                  bool rv =(!error) && (fileStatus.st_mode & _S_IFDIR);

                  if(inverted_)
                     rv = !rv;

                  return rv;
              }

           case FILEOP_EMPTY:
              {
                  if(verbose)
                     cout << (inverted_ ? "!" : "") << "-s " << string_ << ' ';
    
                  bool rv = error || (fileStatus.st_size == 0);

                  if(inverted_)
                     rv = !rv;
                  
                  return rv;
              }

           default:
             cerr << "error:  bug in code, invalid fileop requested" << endl;
             break;
       }

       return false;  // fail by default
    }

    virtual string stringValue() const { return string_; }


private:
    string    string_;
    FILEOP    op_;
};



////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////

struct TestCommand
   /// perform a 'test' operation on a bunch of canned strings.
   ///
   /// The test can use the strings as names of files and thus perform useful file system queries.
   ///
   /// BNF:
   ///
   ///   testCommandSyntax:=    '[' expr ']'
   ///   expr             :=     term [ -o expr ]
   ///   term             :=     primitive [ -a term ]
   ///   primitive        :=     (fileOperation, stringComparison)
   ///   fileOperation    :=     ('-f' fileName, '-r' fileName, '-d' fileName, '-s' fileName)
   ///   stringComparison :=     (string [ binaryOperator string ])
   ///   binaryoperator   :=     (strOp, mathOp)
   ///   strOp            :=     ('>', '<', '=', '>=', '<=')
   ///   mathOp           :=     ('-eq', '-ne', '-ge', '-le')
   ///
   ///
: public Command
{
   TestCommand(TestExpression*e)
   : expression_(e)
   {
   }

   virtual ~TestCommand() {}


   virtual bool execute() const 
   {
      if(verbose)
         cout << "[" << endl;

      bool rv = expression_->execute();

      if(verbose)
         cout << "]" << endl;


      return rv;
   }

private:
    auto_ptr<TestExpression> expression_;
};


template<class Iterator, class constIterator>
void skipBlanks(Iterator &p, constIterator const &e)
{

    while((p != e) && (*p <= ' '))
        ++p;

}


static void tokenize(string const &s, list<string> &tokens)
{

    string::const_iterator f = s.begin(), l = s.end();

    while(f != l)
    {
        skipBlanks(f,l);

        if(f == l)
           break;

        char c = *f; 


        if(c == '"')
        {
           string tmp;

           tmp += c;

           ++f;

           while(f != l && (*f != '"'))
               tmp += *f++;

           tmp += '"';

           if(f != l)
               ++f;  // skip trailing "

           tokens.push_back(tmp);  // includes quotes

        }
        else
        if(c == ';')
        {

            static string semi(";");

            tokens.push_back(semi);

            ++f;
            
        }
        else
        {
            if(c == '\\')
            {
               // handle the possibility of "\ " being in the stream and treat it like
               // space

               ++f;

               if(f == l || *f > ' ')
               {
                 --f;
               }
               else
               {
                  // ignore '\ ' in the stream
                 continue;
               }

            }


            string tmp;

            while((f != l) && (*f > ' '))
            {
                tmp += *f++;       // we should really handle strings like this here:  'stuff"asd adsf  asdf"morestuff' as one token
            }

            tokens.push_back(tmp);
        }
    }

}

Command *parseOneCommand(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken);
Command *parseCommandList(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken);


static string thenToken("then");
static string elseToken("else");
static string fiToken("fi");
static string ifToken("if");
static string whileToken("while");
static string forToken("for");
static string doneToken("done");
static string doToken("do");
static string breakToken("break");
static string continueToken("continue");
static string openSquare("[");
static string closeSquare("]");
static string bangToken("!");


static string eqToken("=");
static string neToken("!=");
static string gtToken("-gt");
static string geToken("-ge");
static string ltToken("-lt");
static string leToken("-le");
static string dashO("-o");
static string dashA("-a");

Command *parseIfCommand(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken)
{
    // assumes that the "if" is already gone.  We are looking for this syntax:
    //
    //  if [!] command parms ; then command list ; [else command list ; ] fi
    //         ^

    bool invert=false;

    if(    (nextToken != lastToken)
       &&  (bangToken == *nextToken)
      )
    {
       invert = true;
       ++nextToken;
    }


    auto_ptr<Command> test(parseOneCommand(nextToken, lastToken));


    if(nextToken == lastToken ||  thenToken != *nextToken)
    {
        cerr << "error: expected 'then' token, not (" << *nextToken << ")" << endl;
        return 0;
    }

    ++nextToken;  // eat 'then'

    auto_ptr<Command> ifClause(parseCommandList(nextToken, lastToken));

    if(    nextToken == lastToken 
       ||  (
                 *nextToken != fiToken
              && *nextToken != elseToken
           )
      )
    {
       cerr << "error:  expected 'fi' or 'else', found: "
            << ((nextToken == lastToken) ? "end of command" : *nextToken)
            << endl;

       return 0;
    }

    auto_ptr<Command> elseClause(0);

    bool badElse = false;


    if(*nextToken == elseToken)
    {
       ++nextToken;

       elseClause.reset(parseCommandList(nextToken, lastToken));

       if(!elseClause.get())
       {
          cerr << "error:  expected command list after else clause" << endl;
          badElse = true;
       }

    }


    if(    badElse
       ||  nextToken == lastToken 
       ||  *nextToken != fiToken
      )
    {
       cerr << "error:  expected 'fi', found: "
            << ( (nextToken == lastToken) ? "end of command" : *nextToken)
            << endl;

       return 0;
    }

    ++nextToken;  // eat 'fi'

    if(   !test.get()
       || !ifClause.get()
      )
      return 0;  // some kind of error occurred
    

    return new IfCommand(test.release(), ifClause.release(), elseClause.release(), invert);

}

TestExpression *parseBinop(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   //   one of these syntaxes:
   //
   //   [ ! -f fred ]
   //       ^
   //
   //   [ a > b ]
   //     ^
   //
   //   [ aString ]
   //     ^
   //


    if(nextToken == lastToken)
    {

       cerr << "error:  expected expression, found emptiness" << endl;

       return 0;
    }


   bool inverted = false;


   if(nextToken != lastToken && *nextToken == bangToken)
   {
       inverted = true;

       ++nextToken;
   }


   string const &firstToken = *nextToken;

   if(!firstToken.empty())
   {
       // we might have a fileop

       static string dashF("-f");
       static string dashR("-r");
       static string dashW("-w");
       static string dashD("-d");
       static string dashX("-x");

       FILEOP op = FILEOP_UNSPEC;

       if(firstToken == dashF)
          op = FILEOP_EXISTS;
       else
       if(firstToken == dashR)
          op = FILEOP_READABLE;
       else
       if(firstToken == dashW)
          op = FILEOP_WRITEABLE;
       else
       if(firstToken == dashD)
          op = FILEOP_ISDIR;
       else
       if(firstToken == dashX)
          op = FILEOP_EXE;

       if(op != FILEOP_UNSPEC)
       {
          list<string>::const_iterator after = nextToken;  

          ++after;

          if(after == lastToken)
          {
            op = FILEOP_UNSPEC; // nothing comes after the fileop so there is a syntax error

            cerr << "error:  expected filename after " << firstToken << endl;

          }
          else
          {
             // we have a first token that could be a fileop, but let us check to see if
             // the token after is one of the comparison operators

             string const &afterToken = *after;

             if(   afterToken == eqToken
                || afterToken == neToken
                || afterToken == gtToken
                || afterToken == ltToken
                || afterToken == geToken
                || afterToken == leToken
               )
             {
                op = FILEOP_UNSPEC;  // we have something like this:  [ -f -lt -g ]
             }

          }
       }


       if(op != FILEOP_UNSPEC)
       {
          // ok, we know know we have something like this:
          // [ -f fred ]
          //   ^

          ++nextToken;  // we've already interpreted it


          auto_ptr<FileOpExpression> e(new FileOpExpression(op, *nextToken++));

          e->setInverted(inverted);

          return e.release();


       }

      
   }  

   // if we get here, we know that the first word in the stream is not a file operation
   //
   //  [ fred = tom ]
   //    ^
   //
   //  firstToken = *nextToken at this point


   ++nextToken;

   if(nextToken == lastToken || *nextToken == closeSquare)
   {
      // we have hit the end of the stream with only the one token



      auto_ptr<StringExpression> e(new StringExpression(firstToken));

      e->setInverted(inverted);

      return e.release();

   }


   string const &secondToken = *nextToken++;

   if(nextToken == lastToken || *nextToken == closeSquare)
   {
      cerr << "error:  expected 2 operands in a binary expression (" 
           << firstToken << ' ' << secondToken 
           << ") but found end of command" << endl;

      return 0; 
   }

   // at this point, the 3'rd token, which is the second operand is in *nextToken

   auto_ptr<TestExpression> e(0);

   if(secondToken == eqToken)
   {
      e.reset(new StringEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == neToken)
   {
      e.reset(new StringNotEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == gtToken)
   {
      e.reset(new StringGreaterExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == ltToken)
   {
      e.reset(new StringLessExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == leToken)
   {
      e.reset(new StringLessEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   if(secondToken == geToken)
   {
      e.reset(new StringGreaterEqualExpression(new StringExpression(firstToken), new StringExpression(*nextToken++)));
   }
   else
   {
       cerr << "error, expected comparision operation, but found (" << secondToken << ")" << endl;

       return 0;
   }

   e->setInverted(inverted);

   return e.release();


}



TestExpression *parseFactor(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   // this syntax:
   //
   //   [ -f fred -a -d bill ]
   //     ^

   auto_ptr<TestExpression> rv(0);


   if(nextToken == lastToken)
   {
      cerr << "Error:  expected 'expr ]', but found nothing." << endl;

      return 0;
   }

   while(nextToken != lastToken)
   {
       if( closeSquare == *nextToken)
       {
           break;
       }


       auto_ptr<TestExpression> exp(parseBinop(nextToken, lastToken));

       if(!exp.get())
       {
          // syntax error, message will already have been printed

          return 0;  // rv will be auto destructed
       }

       if(rv.get())
       {
          // we already had a left hand side and now we are adding the right

          auto_ptr<TestExpression> newbie(new AndExpression(rv.release(), exp.release()));

          // rv is now empty, and newbie has the old rv as it's left hand side, and a newly
          // gathered term expression as it's right hand side

          rv.reset(  newbie.release() );  // newbie is empty, and now rv is populated
       }
       else
       {
          // this the first term in the expression

          rv.reset(exp.release());  // exp is now empty and rv has it's term

       }

       if(nextToken == lastToken || *nextToken != dashA)
           break;

       ++nextToken;  // eat the -a


   }

   if(!rv.get())
   {
      cerr << "error:  missing test expression" << endl;
   }


   return rv.release();
}


TestExpression *parseTerm(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   // this syntax:
   //
   //   [ -f fred -a -d bill ]
   //     ^

   auto_ptr<TestExpression> rv(0);


   if(nextToken == lastToken)
   {
      cerr << "Error:  expected 'expr ]', but found nothing." << endl;

      return 0;
   }


   if(nextToken != lastToken && *nextToken != closeSquare)
   {

       auto_ptr<TestExpression> exp(parseFactor(nextToken, lastToken));

       if(!exp.get())
       {
          // syntax error, message will already have been printed

          return 0;  // rv will be auto destructed
       }

       if(rv.get())
       {
          // we already had a left hand side and now we are adding the right

          auto_ptr<TestExpression> newbie(new AndExpression(rv.release(), exp.release()));

          // rv is now empty, and newbie has the old rv as it's left hand side, and a newly
          // gathered term expression as it's right hand side

          rv.reset(  newbie.release() );  // newbie is empty, and now rv is populated
       }
       else
       {
          // this the first term in the expression

          rv.reset(exp.release());  // exp is now empty and rv has it's term

       }


   }

   if(!rv.get())
   {
      cerr << "error:  missing test expression" << endl;
   }


   return rv.release();
}


TestExpression *parseTestExpression(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken)
{
   // assumes that the leading square bracket has been already discarded from the stream.  We are parsing
   // this syntax:
   //
   //   [ -f fred -o -d bill ]
   //     ^

   auto_ptr<TestExpression> rv(0);

   if(nextToken == lastToken)
   {
      cerr << "Error:  expected 'expr ]', but found nothing." << endl;

      return 0;
   }

   while(nextToken != lastToken)
   {
       if( closeSquare == *nextToken)
       {
           break;
       }

       auto_ptr<TestExpression> exp(parseTerm(nextToken, lastToken));

       if(!exp.get())
       {
          // syntax error, message will already have been printed

          return 0;  // rv will be auto destructed
       }

       if(rv.get())
       {
          // we already had a left hand side and now we are adding the right

          auto_ptr<TestExpression> newbie(new OrExpression(rv.release(), exp.release()));

          // rv is now empty, and newbie has the old rv as it's left hand side, and a newly
          // gathered term expression as it's right hand side

          rv.reset(  newbie.release() );  // newbie is empty, and now rv is populated
       }
       else
       {
          // this the first term in the expression

          rv.reset(exp.release());  // exp is now empty and rv has it's term

       }

       if(nextToken == lastToken || *nextToken != dashO)
          break;  // only keep going if we  currently see -o in the stream

       ++nextToken;  // eat the -o


   }

   if(!rv.get())
   {
      cerr << "error:  missing test expression" << endl;
   }

   if(nextToken != lastToken && *nextToken == closeSquare)
      ++nextToken; // eat the trailing ] in the expresssion

   return rv.release();
}


Command *parseOneCommand(list<string>::const_iterator &nextToken, list<string>::const_iterator const &lastToken)
{
    Command *rv=0;

    
    if(nextToken->empty())
    {
       cerr << "Unexpected empty token!" << endl;
       delete rv;
       return 0;
    }
    
    static string semicolon(";");
    
    if(semicolon == *nextToken)
    {
       ++nextToken;
    
       if(rv == 0)
          rv = new Command(); // empty command
    
       return rv;
    }
    
    // if we get here, we have tokens in the stream and
    // they are not ;
    
    
    string const &firstToken = *nextToken;
    
    if(firstToken == ifToken)
    {
       ++nextToken;  // skip the if

       rv = parseIfCommand(nextToken, lastToken);
    
       // parseIfStatement will print errors as needed
    
    }
    else
    if(firstToken == fiToken || firstToken == doneToken)
    {
        if(rv == 0)
        {
           cerr << "error:  found (" << firstToken << ") but was expecting a statement" << endl;
        }

        return rv;
    }
    else
    if(firstToken == openSquare)
    {
       ++nextToken;

       TestExpression *expression = parseTestExpression(nextToken, lastToken);

	   if( nextToken != lastToken && *nextToken == semicolon)
	       ++nextToken;

       if(expression)
       {
          rv = new TestCommand(expression);
       }
       // else, we will already have produced an error message


    }
    else
    {
       // if we get here, we are doing a normal command
    
       string cmd;
    
       while(    nextToken != lastToken 
             &&  semicolon != *nextToken
            )
       {


           if(!cmd.empty())
           {
              cmd += ' ';
              cmd += *nextToken++;
           }
           else
           {
              // on the first word of the line, convert / to \

              string cmdPath;

              string tmp(*nextToken++);

              string::const_iterator f = tmp.begin(), l = tmp.end();

              while(f != l)
              {
                 char c = *f++;

                 if(c == '/')
                    c = '\\';

                 cmdPath += c;
              }


              cmd += cmdPath;
           }
    
       }


       if(nextToken != lastToken)
          ++nextToken; // eat the trailing ;
    
       rv = new RegularCommand(cmd);
    
    }
    
    return rv;

}



Command *parseCommandList(list<string>::const_iterator &f,
                          list<string>::const_iterator const &l
                          
                         )
    ///  parse a string into a command pointer (or sequence of commands).
    ///  invalid command syntax will produce an error to stderr and a null
    ///  will be returned.
{
    list<string>::const_iterator b = f; // save the whole string for error messages


    list<Command*> commands;

    while(f != l)
    {
       // now that we have a list of tokens, we can parse the commands out of them

       if(   *f == fiToken 
          || *f == doneToken
          || *f == elseToken
         )
       {
          break;
       }


       Command *p = parseOneCommand(f,l);

       if(p)
         commands.push_back(p);
       else
       {
          cerr << "Invalid command structure:  ";

          while(b != l)
             cerr << *b++ << ' ';

          cerr << endl;
          
          list<Command*>::iterator a = commands.begin(), b = commands.end();

          while(a != b)
              delete *a++;

          return 0;

       }
          
    }

    if(commands.empty())
    {
        cerr << "error:  empty command or invalid command type:  " ;
        
        while(b != l)
            cerr << *b++ << ' ';


        cerr << endl;
        return 0;
    }

    if(commands.size() == 1)
    {
       return *commands.begin();
    }

    return new CommandList(commands);
}


Command *parseCommandList(string const &cmd)
    ///  parse a string into a command pointer (or sequence of commands).
    ///  invalid command syntax will produce an error to stderr and a null
    ///  will be returned.
{

    list<string> tokens;

    tokenize(cmd, tokens);


    list<string>::const_iterator f = tokens.begin(), l = tokens.end();


    return parseCommandList(f,l);


}

static void parseAndExecute(string const &cmd)
{
    Command *p = parseCommandList(cmd);

    if(p)
        p->execute();

    delete p;
}


